#!/usr/bin/env sh

set -e

usage() {

	printf 'Usage %s ' "${1}";
	printf '[-h|--help] [-d|--debug] [-n|--no-wiringpi]\n';
	printf '\t\t[-c \033[4mcflags\033[m] ';
	printf '[-l \033[4mldlibs\033[m] ';
	printf '[--ldflags \033[4mldflags\033[m]\n\t\t';
	printf '[-p \033[4mpkgconfig package\033[m] ';
	printf '[--cc \033[4mcompiler\033[m] ';
	printf '\033[4mbuild dir\033[m\n\n';

}

ARGS=$(getopt -o "hdnc:l:p:" \
	--long "help,debug,no-wiringpi,cflags:,ldlibs:,ldflags:,pkg:,cc:" \
	-n "$0" -- "$@");
if [ "$?" -ne "0" ]; then
	exit 1;
fi

eval set -- "$ARGS";

zero="$0";
PKGS="gtk4 librsvg-2.0"

while true; do

	case "$1" in
		'-h'|'--help')
			usage "${zero}";
			exit 0;
			;;

		'-d'|'--debug')
			DEBUG=1;
			shift;
			;;

		'-n'|'--no-wiringpi')
			CFLAGS="${CFLAGS} -DNO_WIRINGPI";
			NO_WIRINGPI=1;
			shift;
			;;

		'-c'|'--cflags')
			CFLAGS="${CFLAGS} ${2}";
			shift 2;
			;;

		'-l'|'--ldlibs')
			LDLIBS="${LDLIBS} ${2}";
			shift 2;
			;;

		'--ldflags')
			LDFLAGS="${LDFLAGS} ${2}";
			shift 2;
			;;

		'-p'|'--pkg')
			case "${PKGS}" in
				'*${2}*') ;;
				*) PKGS="${PKGS} ${2}";;
			esac
			shift 2;
			;;

		'--cc')
			CC="${2}";
			shift 2;
			;;

		'--')
			shift;
			break;
			;;

		*)
			usage "${zero}" >&2;
			exit1 ;
			;;

	esac

done

if [ -z "${1}" ]; then
	usage "${zero}" >&2;
	exit 1;
fi

builddir="$(realpath "${1}")";
mkdir -p "${builddir}";

srcdir="$(realpath --relative-to "${builddir}" "$(dirname "${zero}")")";
srcs="$(cd "${builddir}/${srcdir}"; find ./src -name "*.c")";
ui_files="$(cd "${builddir}/${srcdir}"; find ./res/ -name "*.blp")";
res_files="$(cd "${builddir}/${srcdir}";
	find ./res/ -name "*.gresource.xml")";
other_res_files=$(cd "${builddir}/${srcdir}";
	find ./res -type f -! -name "*.blp" -a -! -name "*.gresource.xml");

if [ -n "${DEBUG}" ]; then
	CFLAGS="${CFLAGS} -DDEBUG -O0 -g3";
else
	CFLAGS="${CFLAGS} -O2";
fi
CFLAGS="${CFLAGS} -Wall -Wextra -Werror";
CFLAGS="${CFLAGS} -Wno-expansion-to-defined";
CFLAGS="${CFLAGS} -D_GNU_SOURCE";
CFLAGS="${CFLAGS} -I${srcdir}/inc -Iinc";
CFLAGS="${CFLAGS} $(pkg-config --cflags ${PKGS})";
LDLIBS="${LDLIBS} -lm $(pkg-config --libs ${PKGS})";
if [ -z "${NO_WIRINGPI}" ]; then
	LDLIBS="${LDLIBS} -lwiringPi"
fi

{

# Blueprint {{{

cat << EOF
rule blueprint
 command = blueprint-compiler compile --output \${out} \${in}
EOF

for ui_file in ${ui_files}; do
	printf 'build %s: blueprint %s\n' \
		"${ui_file%.blp}.ui" "${srcdir}/${ui_file}";
done

#}}}

# Resources {{{

cat << EOF
rule copy
 command=cp \${in} \${out}

EOF

for res in ${other_res_files}; do
	printf 'build %s: copy %s\n' "${res}" "${srcdir}/${res}";
done
printf '\n';

cat << EOF
rule resource
 command = glib-compile-resources \${xml} --target=\${out} \$
  --generate-source

EOF

for res in ${res_files}; do
	printf "build %s: resource %s" "${res%.xml}.c" "${srcdir}/${res}";
	for ui_file in ${ui_files}; do
		printf ' %s' "${ui_file%.blp}.ui";
	done
	for res_file in ${other_res_files}; do
		printf ' %s' "${res_file}";
	done
	printf '\n';
	printf ' xml = %s\n' "${srcdir}/${res}";
done
printf '\n';

#}}}

# Regular old compiling {{{

cat << EOF
cflags = ${CFLAGS}
ldflags = ${LDFLAGS}
ldlibs = ${LDLIBS}
cc = ${CC:-gcc}

rule compile
 deps = gcc
 depfile = \${out}.d
 command = \${cc} -MD -MF \${out}.d \${cflags} -c -o \${out} \${in}
rule link
 command = \${cc} \${ldflags} -o \${out} \${in} \${ldlibs}

EOF

for src in ${srcs}; do
	printf 'build %s: compile %s\n' "${src%.c}.o" "${srcdir}/${src}";
	case ${src} in
		./src/thirdparty/*)
			printf ' cflags = %s -I%s\n' \
				"${CFLAGS}" "${srcdir}/inc/thirdparty";
			;;
		*) ;;
	esac
done

for res in ${res_files}; do
	printf 'build %s: compile %s\n' "${res%.xml}.o" "${res%.xml}.c";
done

printf 'build game: link'
for src in ${srcs}; do
	printf ' %s' "${src%.c}.o";
done
for src in ${ppsrcs}; do
	printf ' %s' "${src%.cpp}.o";
done
for res in ${res_files}; do
	printf ' %s' "${res%.xml}.o";
done
printf '\n\n';

#}}}

# Reconfiguring {{{

cat << EOF
rule configure
 generator = 1
 command = cd ${srcdir}; ./configure.sh ${ARGS}
build build.ninja: configure ${srcdir}/configure.sh

EOF
#}}}

cat << EOF
default game
EOF

} > ${builddir}/build.ninja
