#include "window.h"

#include "pins.h"

#include <assert.h>
#include <librsvg/rsvg.h>
#include <pango/pangocairo.h>
#ifndef NO_WIRINGPI
#include <wiringPi.h>
#endif

/* TYPES!!! */

struct button_state {
	GameWindow* window;
	bool bouncing;
	bool pressed;
};

struct _GameWindow {

	GtkWindow parent;

	GtkDrawingArea* area;

	RsvgHandle* logo;

	enum {
		SPLASH,
		WAIT,
		LIT,
		DONE,
	} state;

	struct {
		int team_1;
		int team_2;
	} score;
	struct {
		bool team_1;
		bool team_2;
	} wrong;

	enum { RED, GREEN } colour;

	struct button_state g1, r1, g2, r2;
	bool g1_bouncing;
	bool r1_bouncing;
	bool g2_bouncing;
	bool r2_bouncing;

};

G_DEFINE_TYPE(
	GameWindow, game_window, GTK_TYPE_WINDOW );

enum colour {
	WHITE, UNIVERSITY_BLUE,
};

/* PROTOTYPES!!! */

static void game_window_class_init( GameWindowClass* class );
static void game_window_init( GameWindow* window );

static void realise( GtkWidget* widget, void* data );

static void handle_press( GameWindow* window, int pin );

static void draw( GtkDrawingArea* area, cairo_t* cr,
	int width, int height, void* data );

static void set_cario_colour( cairo_t* cr, enum colour colour );

static int bounce_reset( void* data );
#ifndef NO_WIRINGPI
static int light( void* data );
#endif

/* PUBLIC FUNK!!! */

GameWindow* game_window_new( GameApp* app ) {

	return g_object_new( GAME_WINDOW_TYPE,
		"application", app, NULL );

}

void game_window_button_press( GameWindow* window, int pin ) {

	fprintf( stderr, "Pressed %d", pin );

	struct button_state* button_state = NULL;

	switch ( pin ) {

		case BUTTON_G1:
			button_state = &window->g1;
			break;

		case BUTTON_R1:
			button_state = &window->r1;
			break;

		case BUTTON_G2:
			button_state = &window->g2;
			break;

		case BUTTON_R2:
			button_state = &window->r2;
			break;

	}

	assert( NULL != button_state );

	if ( button_state->bouncing ) {
		fprintf( stderr, "(bouncing)\n" );
		return;
	} else {
		fprintf( stderr, "\n" );
	}

	if ( button_state->pressed ) {
		return;
	}

	fprintf( stderr, "Ping %d\n", pin );

	button_state->bouncing = true;
	button_state->pressed = true;

	g_timeout_add( 200, bounce_reset, button_state );

	handle_press( window, pin );

}

void game_window_button_release( GameWindow* window, int pin ) {

	fprintf( stderr, "Rele'ed %d", pin );

	struct button_state* button_state = NULL;

	switch ( pin ) {

		case BUTTON_G1:
			button_state = &window->g1;
			break;

		case BUTTON_R1:
			button_state = &window->r1;
			break;

		case BUTTON_G2:
			button_state = &window->g2;
			break;

		case BUTTON_R2:
			button_state = &window->r2;
			break;

	}

	assert( NULL != button_state );

	if ( button_state->bouncing ) {
		fprintf( stderr, "(bouncing)\n" );
		return;
	} else {
		fprintf( stderr, "\n" );
	}

	if ( !button_state->pressed ) {
		return;
	}

	button_state->bouncing = true;
	button_state->pressed = false;

	g_timeout_add( 200, bounce_reset, button_state );

}

/* PRIVATE FUNK!!! */

static void game_window_class_init( GameWindowClass* class ) {

	gtk_widget_class_set_template_from_resource(
		GTK_WIDGET_CLASS( class ), "/xyz/tomg/game/res/window.ui" );
	gtk_widget_class_bind_template_child(
		GTK_WIDGET_CLASS( class ), GameWindow, area );

	gtk_widget_class_bind_template_callback_full(
		GTK_WIDGET_CLASS( class ), "realise",
		G_CALLBACK( &realise ) );

}

static void game_window_init( GameWindow* window ) {

	gtk_widget_init_template( GTK_WIDGET( window ) );

	GBytes* logo_svg = g_resources_lookup_data(
		"/xyz/tomg/game/res/logo.svg", 0, NULL );
	GError* error = NULL;
	const char* data = g_bytes_get_data( logo_svg, NULL );
	window->logo = rsvg_handle_new_from_data(
		(const unsigned char*) data, strlen( data ), &error );
	if ( NULL != error ) {
		fprintf( stderr, "%s\n", error->message );
	}
	g_bytes_unref( logo_svg );

	window->state = SPLASH;

	window->g1.window = window;
	window->r1.window = window;
	window->g2.window = window;
	window->r2.window = window;

	gtk_drawing_area_set_draw_func(
		window->area, &draw, (void*) window, NULL );

	gtk_widget_set_cursor( GTK_WIDGET( window ),
		gdk_cursor_new_from_name( "none", NULL ) );

}

static void realise( GtkWidget* widget, void* data ) {

	gtk_window_fullscreen( GTK_WINDOW( widget ) );
	(void) data;

}

static void handle_press( GameWindow* window, int pin ) {

#ifndef NO_WIRINGPI

	if ( !digitalRead( BUTTON_G1 ) && !digitalRead( BUTTON_R1 ) &&
			!digitalRead( BUTTON_G2 ) && !digitalRead( BUTTON_R2 ) ) {

		fprintf( stderr, "All pressed, reset\n" );

		window->state = SPLASH;
		window->score.team_1 = 0;
		window->score.team_2 = 0;
		gtk_widget_queue_draw( GTK_WIDGET( window->area ) );
		return;

	}

	switch ( window->state ) {

		case SPLASH: {

			fprintf( stderr, "Splash -> wait\n" );

			window->state = WAIT;
			window->score.team_1 = 0;
			window->score.team_2 = 0;
			g_timeout_add( g_random_int_range( 1000, 5000 ),
				light, window );
			break;

		}

		case WAIT: {

			switch ( pin ) {

				case BUTTON_G1:
				case BUTTON_R1: {

					if ( window->score.team_1 > 0 &&
							!window->wrong.team_1 ) {
						window->score.team_1--;
						window->wrong.team_1 = true;
					}

					break;

				}

				case BUTTON_G2:
				case BUTTON_R2: {

					if ( window->score.team_2 > 0 &&
							!window->wrong.team_2 ) {
						window->score.team_2--;
						window->wrong.team_2 = true;
					}

					break;

				}

			}

			break;
		}

		case LIT: {

			switch ( pin ) {

				case BUTTON_G1:
				case BUTTON_R1: {

					fprintf( stderr, "Team 1 press, " );
					if ( ( GREEN == window->colour ) ==
							( BUTTON_G1 == pin ) ) {
						fprintf( stderr, "correct\n" );
						window->score.team_1++;
						digitalWrite( LED_G1, LOW );
						digitalWrite( LED_R1, LOW );
						digitalWrite( LED_G2, LOW );
						digitalWrite( LED_R2, LOW );
						digitalWrite( LED_G3, LOW );
						digitalWrite( LED_R3, LOW );
						digitalWrite( LED_G4, LOW );
						digitalWrite( LED_R4, LOW );
						window->wrong.team_1 = false;
						window->wrong.team_2 = false;

						window->state = WAIT;

					} else {
						fprintf( stderr, "wrong\n" );
						if ( window->score.team_1 > 0 &&
								!window->wrong.team_1 ) {
							window->score.team_1--;
							window->wrong.team_1 = true;
						}
					}

					break;

				}

				case BUTTON_G2:
				case BUTTON_R2: {

					fprintf( stderr, "Team 2 press, " );
					if ( ( GREEN == window->colour ) ==
							( BUTTON_G2 == pin ) ) {

						fprintf( stderr, "correct\n" );
						window->score.team_2++;
						digitalWrite( LED_G1, LOW );
						digitalWrite( LED_R1, LOW );
						digitalWrite( LED_G2, LOW );
						digitalWrite( LED_R2, LOW );
						digitalWrite( LED_G3, LOW );
						digitalWrite( LED_R3, LOW );
						digitalWrite( LED_G4, LOW );
						digitalWrite( LED_R4, LOW );
						window->wrong.team_1 = false;
						window->wrong.team_2 = false;

						window->state = WAIT;

					} else {
						fprintf( stderr, "wrong\n" );
						if ( window->score.team_2 > 0 &&
								!window->wrong.team_2 ) {
							window->score.team_2--;
							window->wrong.team_2 = true;
						}
					}

					break;

				}

			}


			if ( window->score.team_2 >= 10 ||
					window->score.team_1 >= 10 ) {

				window->state = DONE;

			}

			if ( window->state == WAIT ) {

				g_timeout_add( g_random_int_range( 1000, 5000 ),
					light, window );

			}

			break;

		}

		case DONE: {

			break;

		}

	}

#else

	(void) pin;

#endif

	gtk_widget_queue_draw( GTK_WIDGET( window->area ) );

}

static void draw( GtkDrawingArea* area, cairo_t* cr,
		int width, int height, void* data ) {

	(void) area;
	GameWindow* window = (GameWindow*) data;

	set_cario_colour( cr, UNIVERSITY_BLUE );
	cairo_paint( cr );

	PangoFontDescription* desc = pango_font_description_new();
	pango_font_description_set_family( desc, "Roboto" );
	pango_font_description_set_absolute_size( desc, PANGO_SCALE * (
		0.8 * height < 0.66 * width ? 0.8 * height : 0.66 * width ) );

	switch ( window->state ) {

		case SPLASH: {

			rsvg_handle_render_document( window->logo,
				cr, &(RsvgRectangle) { 0, 0, width, height, }, NULL );
			break;

		}

		case WAIT:
		case LIT: {

			char buffer[5];
			snprintf( buffer, sizeof buffer,
				"%d", window->score.team_1 );

			int lw, lh;

			PangoLayout* layout = pango_cairo_create_layout( cr );
			pango_layout_set_font_description( layout, desc );

			pango_layout_set_text( layout, buffer, -1 );
			pango_layout_get_pixel_size( layout, &lw, &lh );

			cairo_move_to( cr, 0.25 * width - lw / 2.,
				0.5 * height - lh / 2. );
			set_cario_colour( cr, WHITE );
			pango_cairo_show_layout( cr, layout );

			snprintf( buffer, sizeof buffer,
				"%d", window->score.team_2 );

			pango_layout_set_text( layout, buffer, -1 );
			pango_layout_get_pixel_size( layout, &lw, &lh );

			cairo_move_to( cr, 0.75 * width - lw / 2.,
				0.5 * height - lh / 2. );
			set_cario_colour( cr, WHITE );
			pango_cairo_show_layout( cr, layout );

			g_object_unref( G_OBJECT( layout ) );

			rsvg_handle_render_document( window->logo, cr,
				&(RsvgRectangle) {
					0.4 * width, 0.8 * height,
					0.2 * width, 0.2 * height,
				}, NULL );

			break;

		}

		case DONE: {

			double size = ( 0.4 * height < 0.33 * width ?
					0.2 * height : 0.16 * width );
			size = 100;
			pango_font_description_set_absolute_size(
				desc, PANGO_SCALE * size );

			PangoLayout* layout = pango_cairo_create_layout( cr );
			pango_layout_set_font_description( layout, desc );

			pango_layout_set_text( layout,
				window->score.team_1 >= window->score.team_2 ?
					"Team 1 Wins!" : "Team 2 Wins!", -1 );
			pango_layout_set_width( layout, PANGO_SCALE * 0.8 * width );
			pango_layout_set_alignment( layout, PANGO_ALIGN_CENTER );

			int lw, lh;
			pango_layout_get_pixel_size( layout, &lw, &lh );

			set_cario_colour( cr, WHITE );
			cairo_move_to( cr,
				0.1 * width, ( height - lh ) / 2 );
			pango_cairo_show_layout( cr, layout );

			break;

		}

	}

	pango_font_description_free( desc );

}

static void set_cario_colour( cairo_t* cr, enum colour colour ) {

	switch ( colour ) {

		case WHITE:
			cairo_set_source_rgba( cr, 1., 1., 1., 1. );
			break;

		case UNIVERSITY_BLUE:
			cairo_set_source_rgba(
				cr, 0, 0x5C / 255., 0x84 / 255., 1. );
			break;

	}

}

static int bounce_reset( void* data ) {

#ifndef NO_WIRINGPI

	struct button_state* state = (struct button_state*) data;

	state->bouncing = false;

	int button = -1;
	if ( state == &( state->window->g1 ) ) {
		button = BUTTON_G1;
	} else if ( state == &( state->window->r1 ) ) {
		button = BUTTON_R1;
	} else if ( state == &( state->window->g2 ) ) {
		button = BUTTON_G2;
	} else if ( state == &( state->window->r2 ) ) {
		button = BUTTON_R2;
	}
	assert( -1 != button );

	bool pressed = ( LOW == digitalRead( button ) );

	fprintf( stderr, "Done bouncing: %s",
		pressed ? "pressed" : "released" );

	if ( pressed != state->pressed ) {

		fprintf( stderr, " (adjusting)\n" );

		state->pressed = pressed;
		if ( pressed ) {
			handle_press( state->window, button );
		}

	} else {
		fprintf( stderr, "\n" );
	}

#else

	(void) data;

#endif

	return G_SOURCE_REMOVE;

}

#ifndef NO_WIRINGPI
static int light( void* data ) {

	GameWindow* window = (GameWindow*) data;

	if ( WAIT != window->state ) {
		return G_SOURCE_REMOVE;
	}

	static const int leds[] = {
		LED_G1, LED_G2, LED_G3, LED_G4,
		LED_R1, LED_R2, LED_R3, LED_R4,
	};

	int led = g_random_int_range( 0, sizeof leds / sizeof *leds );
	fprintf( stderr, "Lighting led %d (pin %d)\n",
		led, leds[led] );
	for ( int i = 0; i < 8; i++ ) {
		digitalWrite( leds[i], i == led ? HIGH : LOW );
	}

	window->state = LIT;
	window->colour = led < 4 ? GREEN : RED;

	return G_SOURCE_REMOVE;

}
#endif
